# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import (
    SpectrumUncorrelatedException,
    determine_shift_by_correlation,
)


class FastAutolock:
    """Spectroscopy autolock based on correlation."""

    def __init__(
        self,
        control,
        parameters,
        first_error_signal,
        first_error_signal_rolled,
        x0,
        x1,
        additional_spectra=None,
    ):
        self.control = control
        self.parameters = parameters

        self.first_error_signal_rolled = first_error_signal_rolled

        self._done = False
        self._error_counter = 0

    def handle_new_spectrum(self, spectrum):
        if self._done:
            return

        try:
            shift, zoomed_ref, zoomed_err = determine_shift_by_correlation(
                1, self.first_error_signal_rolled, spectrum
            )
        except SpectrumUncorrelatedException:
            self._error_counter += 1
            print("skipping spectrum because it is not correlated")

            if self._error_counter > 10:
                raise

            return

        lock_point = int(
            round((shift * (-1)) * self.parameters.sweep_amplitude.value * 8191)
        )

        print("lock point is", lock_point, shift)

        self.parameters.autolock_target_position.value = int(lock_point)
        self.parameters.autolock_preparing.value = False
        self.control.exposed_write_registers()
        self.control.exposed_start_lock()

        self._done = True

    def after_lock(self):
        pass



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from time import time

import numpy as np
from linien_common.common import (
    AUTOLOCK_MAX_N_INSTRUCTIONS,
    SpectrumUncorrelatedException,
    determine_shift_by_correlation,
)
from linien_server.autolock.utils import (
    crop_spectra_to_same_view,
    get_all_peaks,
    get_diff_at_time_scale,
    get_lock_region,
    get_target_peak,
    get_time_scale,
    sign,
    sum_up_spectrum,
)


class LockPositionNotFound(Exception):
    pass


class UnableToFindDescription(Exception):
    pass


class RobustAutolock:
    def __init__(
        self,
        control,
        parameters,
        first_error_signal,
        first_error_signal_rolled,
        x0,
        x1,
        N_spectra_required=5,
        additional_spectra=None,
    ):
        self.control = control
        self.parameters = parameters

        self.first_error_signal = first_error_signal
        self.x0 = x0
        self.x1 = x1

        self.N_spectra_required = N_spectra_required

        self.spectra = [first_error_signal]

        self._done = False
        self._error_counter = 0

        if additional_spectra is not None:
            # the most recent ones are the ones we are interested in
            additional_spectra = list(reversed(additional_spectra))
            for additional_spectrum in additional_spectra:
                self.handle_new_spectrum(additional_spectrum)

    def handle_new_spectrum(self, spectrum):
        if self._done:
            return

        print("handle new spectrum")
        try:
            determine_shift_by_correlation(1, self.first_error_signal, spectrum)
        except SpectrumUncorrelatedException:
            print("skipping spectrum because it is not correlated")
            self._error_counter += 1
            if self._error_counter > 2:
                raise

            return

        self.spectra.append(spectrum)
        self.parameters.autolock_percentage.value = int(
            round((len(self.spectra) / self.N_spectra_required) * 100)
        )

        if len(self.spectra) == self.N_spectra_required:
            print("enough spectra!, calculate")

            t1 = time()
            description, final_wait_time, time_scale = calculate_autolock_instructions(
                self.spectra, (self.x0, self.x1)
            )
            t2 = time()
            print("calculation of autolock description took", t2 - t1)

            # sets up a timeout: if the lock doesn't finish within a certain time
            # span, throw an error
            self.setup_timeout()

            # first reset lock in case it was True. This ensures that autolock
            # starts properly once all parameters are set
            self.parameters.lock.value = False
            self.control.exposed_write_registers()

            self.parameters.autolock_time_scale.value = time_scale
            self.parameters.autolock_instructions.value = description
            self.parameters.autolock_final_wait_time.value = final_wait_time

            self.control.exposed_write_registers()

            self.parameters.lock.value = True
            self.control.exposed_write_registers()

            self.parameters.autolock_preparing.value = False

            self._done = True

        else:
            print(
                "not enough spectra collected: %d of %d"
                % (len(self.spectra), self.N_spectra_required)
            )

    def setup_timeout(self, N_acquisitions_to_wait=5):
        """Robust autolock just programs the FPGA image with a set of instructions.
        The FPGA image then uses these instructions in order to actually turn on
        the lock once all conditions are met. However, it may happen that the
        FPGA image is unable to lock for some reason. For this case, we set up
        a timeout that raises an error if this happens.
        """
        self._timeout_start_time = time()
        self._timeout_time_to_wait = (
            N_acquisitions_to_wait
            * 2
            * sweep_speed_to_time(self.parameters.sweep_speed.value)
        )

        self.parameters.ping.add_callback(self.check_for_timeout)

    def check_for_timeout(self, ping):
        min_time_to_wait = 5

        if time() - self._timeout_start_time > max(
            self._timeout_time_to_wait, min_time_to_wait
        ):
            print("Waited too long for autolock! Aborting")
            self.stop_timeout()
            self.parameters.task.value.exposed_stop()

    def stop_timeout(self):
        self.parameters.ping.remove_callback(self.check_for_timeout)

    def after_lock(self):
        self.stop_timeout()


def calculate_autolock_instructions(spectra_with_jitter, target_idxs):
    spectra, crop_left = crop_spectra_to_same_view(spectra_with_jitter)

    target_idxs = [idx - crop_left for idx in target_idxs]

    time_scale = int(
        round(np.mean([get_time_scale(spectrum, target_idxs) for spectrum in spectra]))
    )

    print("x scale is %d" % time_scale)

    prepared_spectrum = get_diff_at_time_scale(sum_up_spectrum(spectra[0]), time_scale)
    peaks = get_all_peaks(prepared_spectrum, target_idxs)
    y_scale = peaks[0][1]

    lock_regions = [get_lock_region(spectrum, target_idxs) for spectrum in spectra]

    for tolerance_factor in [0.95, 0.9, 0.85, 0.8, 0.75, 0.7, 0.65, 0.6, 0.55, 0.5]:
        print("try out tolerance", tolerance_factor)
        peaks_filtered = [
            (peak_position, peak_height * tolerance_factor)
            for peak_position, peak_height in peaks
        ]
        # it is important to do the filtering that happens here after the previous
        # line as the previous line shrinks the values
        peaks_filtered = [
            (peak_position, peak_height)
            for peak_position, peak_height in peaks_filtered
            if abs(peak_height) > abs(y_scale * (1 - tolerance_factor))
        ]

        # now find out how much we have to wait in the end (because we detect the peak
        # too early because our threshold is too low)
        target_peak_described_height = peaks_filtered[0][1]
        target_peak_idx = get_target_peak(prepared_spectrum, target_idxs)
        current_idx = target_peak_idx
        while True:
            current_idx -= 1
            if np.abs(prepared_spectrum[current_idx]) < np.abs(
                target_peak_described_height
            ):
                break
        final_wait_time = target_peak_idx - current_idx
        print("final wait time is %d samples" % final_wait_time)

        description = []

        last_peak_position = 0
        for peak_position, peak_height in list(reversed(peaks_filtered)):
            # TODO: this .9 factor is very arbitrary.
            description.append(
                (int(0.9 * (peak_position - last_peak_position)), int(peak_height))
            )
            last_peak_position = peak_position

        # test whether description works fine for every recorded spectrum
        does_work = True
        for spectrum, lock_region in zip(spectra, lock_regions):
            try:
                lock_position = get_lock_position_from_autolock_instructions(
                    spectrum, description, time_scale, spectra[0], final_wait_time
                )
                if not lock_region[0] <= lock_position <= lock_region[1]:
                    raise LockPositionNotFound()

            except LockPositionNotFound:
                does_work = False

        if does_work:
            break
    else:
        raise UnableToFindDescription()

    if len(description) > AUTOLOCK_MAX_N_INSTRUCTIONS:
        print("warning: autolock description too long. Cropping!", len(description))
        description = description[-AUTOLOCK_MAX_N_INSTRUCTIONS:]

    print("description is", description)
    return description, final_wait_time, time_scale


def get_lock_position_from_autolock_instructions(
    spectrum, description, time_scale, initial_spectrum, final_wait_time
):
    summed = sum_up_spectrum(spectrum)
    summed_xscaled = get_diff_at_time_scale(summed, time_scale)

    description_idx = 0

    last_detected_peak = 0

    for idx, value in enumerate(summed_xscaled):
        wait_for, current_threshold = description[description_idx]

        if (
            sign(value) == sign(current_threshold)
            and abs(value) >= abs(current_threshold)
            and idx - last_detected_peak > wait_for
        ):
            description_idx += 1
            last_detected_peak = idx

            if description_idx == len(description):
                # this was the last peak!
                return idx + final_wait_time

    raise LockPositionNotFound()


def sweep_speed_to_time(sweep_speed):
    """Sweep speed is an arbitrary unit (cf. `parameters.py`).
    This function converts it to the duration of the sweep in seconds.
    """
    f_real = 3.8e3 / (2**sweep_speed)
    duration = 1 / f_real
    return duration



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import _thread
import atexit
import pickle
from copy import copy
from random import randint, random
from threading import Event, Thread
from time import sleep
from typing import List, Tuple

import click
import numpy as np
import rpyc
from linien_common.common import N_POINTS, check_plot_data, update_signal_history
from linien_common.communication import pack, unpack
from linien_common.config import DEFAULT_SERVER_PORT
from linien_common.influxdb import InfluxDBCredentials, restore_credentials
from linien_server import __version__
from linien_server.autolock.autolock import Autolock
from linien_server.influxdb import InfluxDBLogger
from linien_server.optimization.optimization import OptimizeSpectroscopy
from linien_server.parameters import Parameters, restore_parameters, save_parameters
from linien_server.pid_optimization.pid_optimization import (
    PIDOptimization,
    PSDAcquisition,
)
from linien_server.registers import Registers
from rpyc.utils.server import ThreadedServer


class BaseService(rpyc.Service):
    """
    A service that provides functionality for seamless integration of parameter access
    on the client.
    """

    def __init__(self) -> None:
        self.parameters = Parameters()
        self.parameters = restore_parameters(self.parameters)
        atexit.register(save_parameters, self.parameters)
        self._uuid_mapping = {}  # type: ignore[var-annotated]

        influxdb_credentials = restore_credentials()
        self.influxdb_logger = InfluxDBLogger(influxdb_credentials, self.parameters)

        self.stop_event = Event()
        self.stop_log_event = Event()

    def on_connect(self, client) -> None:
        self._uuid_mapping[client] = client.root.uuid

    def on_disconnect(self, client) -> None:
        uuid = self._uuid_mapping[client]
        self.parameters.unregister_remote_listeners(uuid)

    def exposed_get_server_version(self) -> str:
        return __version__

    def exposed_get_param(self, param_name: str) -> bytes:
        return pack(getattr(self.parameters, param_name).value)

    def exposed_set_param(self, param_name: str, value: bytes) -> None:
        getattr(self.parameters, param_name).value = unpack(value)

    def exposed_reset_param(self, param_name: str) -> None:
        getattr(self.parameters, param_name).reset()

    def exposed_init_parameter_sync(self, uuid: str) -> bytes:
        return pack(list(self.parameters.init_parameter_sync(uuid)))

    def exposed_register_remote_listener(self, uuid: str, param_name: str) -> None:
        self.parameters.register_remote_listener(uuid, param_name)

    def exposed_register_remote_listeners(
        self, uuid: str, param_names: List[str]
    ) -> None:
        for param_name in param_names:
            self.exposed_register_remote_listener(uuid, param_name)

    def exposed_get_changed_parameters_queue(self, uuid: str) -> bytes:
        return pack(self.parameters.get_changed_parameters_queue(uuid))

    def exposed_set_parameter_log(self, param_name: str, value: bool) -> None:
        if getattr(self.parameters, param_name).log != value:
            print("Setting log for %s to %s" % (param_name, value))
            getattr(self.parameters, param_name).log = value

    def exposed_get_parameter_log(self, param_name: str) -> bool:
        return getattr(self.parameters, param_name).log

    def exposed_update_influxdb_credentials(
        self, credentials: InfluxDBCredentials
    ) -> Tuple[bool, int, str]:
        credentials = copy(credentials)
        (
            connection_succesful,
            status_code,
            message,
        ) = self.influxdb_logger.test_connection(credentials)
        if connection_succesful:
            self.influxdb_logger.credentials = credentials
            print("InfluxDB credentials updated successfully")
        else:
            print(
                "InfluxDB credentials update failed. Error message: %s (Status Code %s)"
                % (message, status_code)
            )
        return connection_succesful, status_code, message

    def exposed_get_influxdb_credentials(self) -> bytes:
        return pack(self.influxdb_logger.credentials)

    def exposed_start_logging(self, interval: float) -> None:
        print("Starting logging")
        self.influxdb_logger.start_logging(interval)

    def exposed_stop_logging(self) -> None:
        print("Stopping logging")
        self.influxdb_logger.stop_logging()

    def exposed_get_logging_status(self) -> bool:
        return not self.influxdb_logger.stop_event.is_set()


class RedPitayaControlService(BaseService):
    """Control server that runs on the RP that provides high-level methods."""

    def __init__(self, host=None):
        self._cached_data = {}
        self.exposed_is_locked = None

        super(RedPitayaControlService, self).__init__()

        self.registers = Registers(control=self, parameters=self.parameters, host=host)
        # Connect the acquisition loop to the parameters: Every received value is pushed
        # to `parameters.to_plot`.
        self.exposed_pause_acquisition()
        self.exposed_continue_acquisition()

        # Start a timer that increases the `ping` parameter once per second. Its purpose
        # is to allow for periodic tasks on the server: just register an `on_change`
        # listener for this parameter.

        self.ping_thread = Thread(
            target=self._send_ping_loop, args=(self.stop_event,), daemon=True
        )
        self.data_pusher_thread = Thread(
            target=self._push_acquired_data_to_parameters,
            args=(self.stop_event,),
            daemon=True,
        )

        self.ping_thread.start()
        self.data_pusher_thread.start()

        self.exposed_write_registers()

    def _send_ping_loop(self, stop_event: Event):
        while not stop_event.is_set():
            self.parameters.ping.value += 1
            if self.parameters.ping.value < 10:
                print("ping", self.parameters.ping.value)
                if self.parameters.ping.value == 9:
                    print("further pings will be suppressed")
            sleep(1)

    def _push_acquired_data_to_parameters(self, stop_event: Event):
        last_hash = None
        while not stop_event.is_set():
            (
                new_data_returned,
                new_hash,
                data_was_raw,
                new_data,
                data_uuid,
            ) = self.registers.acquisition.exposed_return_data(last_hash)
            if new_data_returned:
                last_hash = new_hash
            # When a parameter is changed, `pause_acquisition` is set. This means that
            # the we should skip new data until we are sure that it was recorded with
            # the new settings.
            if not self.parameters.pause_acquisition.value:
                if data_uuid != self.data_uuid:
                    continue

                data_loaded = pickle.loads(new_data)

                if not data_was_raw:
                    is_locked = self.parameters.lock.value

                    if not check_plot_data(is_locked, data_loaded):
                        print("incorrect data received for lock state, ignoring!")
                        continue

                    self.parameters.to_plot.value = new_data

                    # generate signal stats
                    stats = {}
                    for signal_name, signal in data_loaded.items():
                        stats["%s_mean" % signal_name] = np.mean(signal)
                        stats["%s_std" % signal_name] = np.std(signal)
                        stats["%s_max" % signal_name] = np.max(signal)
                        stats["%s_min" % signal_name] = np.min(signal)
                    self.parameters.signal_stats.value = stats
                    # update signal history (if in locked state)
                    (
                        self.parameters.control_signal_history.value,
                        self.parameters.monitor_signal_history.value,
                    ) = update_signal_history(
                        self.parameters.control_signal_history.value,
                        self.parameters.monitor_signal_history.value,
                        data_loaded,
                        is_locked,
                        self.parameters.control_signal_history_length.value,
                    )
                else:
                    self.parameters.acquisition_raw_data.value = new_data
            sleep(0.05)

    def _task_running(self):
        return (
            self.parameters.autolock_running.value
            or self.parameters.optimization_running.value
            or self.parameters.psd_acquisition_running.value
            or self.parameters.psd_optimization_running.value
        )

    def exposed_write_registers(self) -> None:
        """Sync the parameters with the FPGA registers."""
        self.registers.write_registers()

    def exposed_start_autolock(self, x0, x1, spectrum, additional_spectra=None):
        spectrum = pickle.loads(spectrum)
        # start_watching = self.parameters.watch_lock.value
        start_watching = False
        auto_offset = self.parameters.autolock_determine_offset.value

        if not self._task_running():
            autolock = Autolock(self, self.parameters)
            self.parameters.task.value = autolock
            autolock.run(
                x0,
                x1,
                spectrum,
                should_watch_lock=start_watching,
                auto_offset=auto_offset,
                additional_spectra=pickle.loads(additional_spectra)
                if additional_spectra is not None
                else None,
            )

    def exposed_start_optimization(self, x0, x1, spectrum):
        if not self._task_running():
            optim = OptimizeSpectroscopy(self, self.parameters)
            self.parameters.task.value = optim
            optim.run(x0, x1, spectrum)

    def exposed_start_psd_acquisition(self):
        if not self._task_running():
            self.parameters.task.value = PSDAcquisition(self, self.parameters)
            self.parameters.task.value.run()

        print("Logging thread stopped.")

    def exposed_start_pid_optimization(self):
        if not self._task_running():
            self.parameters.task.value = PIDOptimization(self, self.parameters)
            self.parameters.task.value.run()

    def exposed_start_sweep(self):
        self.exposed_pause_acquisition()
        self.parameters.combined_offset.value = 0
        self.parameters.lock.value = False
        self.exposed_write_registers()
        self.exposed_continue_acquisition()

    def exposed_start_lock(self):
        self.exposed_pause_acquisition()
        self.parameters.lock.value = True
        self.exposed_write_registers()
        self.exposed_continue_acquisition()

    def exposed_shutdown(self):
        self.stop_event.set()
        self.ping_thread.join()
        self.data_pusher_thread.join()
        self.registers.acquisition.exposed_stop_acquisition()
        # FIXME: hacky way to trigger atexit handlers for saving parameters
        _thread.interrupt_main()
        raise SystemExit()

    def exposed_pause_acquisition(self):
        """
        Pause continuous acquisition. Call this before changing a parameter that alters
        the error / control signal. This way, no inconsistent signals reach the
        application. After setting the new parameter values, call
        `continue_acquisition`.
        """
        self.parameters.pause_acquisition.value = True
        self.data_uuid = random()
        self.registers.acquisition.exposed_pause_acquisition()

    def exposed_continue_acquisition(self):
        """
        Continue acquisition after a short delay, when we are sure that the new
        parameters values have been written to the FPGA and that data that is now
        recorded is recorded with the correct parameters.
        """
        self.parameters.pause_acquisition.value = False
        self.registers.acquisition.exposed_continue_acquisition(self.data_uuid)

    def exposed_set_csr_direct(self, key: str, value: int) -> None:
        """
        Directly sets a CSR register. This method is intended for debugging. Normally,
        the FPGA should be controlled via manipulation of parameters.
        """
        self.registers.set(key, value)


class FakeRedPitayaControlService(BaseService):
    def __init__(self):
        super().__init__()
        self.exposed_is_locked = None

        self.random_data_thread = Thread(
            target=self._write_random_data_to_parameters_loop,
            args=(self.stop_event,),
            daemon=True,
        )
        self.random_data_thread.start()

    def _write_random_data_to_parameters_loop(self, stop_event: Event):
        while not stop_event.is_set():
            max_ = randint(0, 8191)
            gen = lambda: np.array([randint(-max_, max_) for _ in range(N_POINTS)])
            self.parameters.to_plot.value = pickle.dumps(
                {
                    "error_signal_1": gen(),
                    "error_signal_1_quadrature": gen(),
                    "error_signal_2": gen(),
                    "error_signal_2_quadrature": gen(),
                }
            )
            sleep(0.1)

    def exposed_write_registers(self):
        pass

    def exposed_start_autolock(self, x0, x1, spectrum):
        print("start autolock", x0, x1)

    def exposed_start_optimization(self, x0, x1, spectrum):
        print("start optimization")
        self.parameters.optimization_running.value = True

    def exposed_shutdown(self):
        raise SystemExit()

    def exposed_pause_acquisition(self):
        pass

    def exposed_continue_acquisition(self):
        pass


@click.command()
@click.version_option(__version__)
@click.argument("port", default=DEFAULT_SERVER_PORT, type=int, required=False)
@click.option(
    "--fake", is_flag=True, help="Runs a fake server that just returns random data"
)
@click.option(
    "--host",
    help=(
        "Allows to run the server locally for development and connects to a RedPitaya. "
        "Specify the RP's host as follows: --host=rp-f0xxxx.local"
    ),
)
def run_server(port, fake=False, host=None):
    print("Start server on port", port)

    if fake:
        print("starting fake server")
        control = FakeRedPitayaControlService()
    else:
        control = RedPitayaControlService(host=host)

    thread = ThreadedServer(
        control,
        port=port,
        protocol_config={"allow_pickle": True},
    )
    thread.start()


if __name__ == "__main__":
    run_server()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
# Copyright 2022 Christian Freier <christian.freier@nomadatomics.com>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import numpy as np
import rpyc
from linien_common.common import (
    HIGH_PASS_FILTER,
    LOW_PASS_FILTER,
    MHz,
    convert_channel_mixing_value,
)
from linien_common.config import ACQUISITION_PORT, DEFAULT_SWEEP_SPEED
from linien_server.parameters import Parameters

from . import csrmap
from .iir_coeffs import make_filter


class Registers:
    """
    This class provides low-level access to the FPGA registers.

    High-level applications should not access this class directly but instead
    communicate by manipulating `Parameters` / `RemoteParameters`.
    """

    def __init__(
        self,
        control,
        parameters: Parameters,
        host: Optional[str] = None,
    ):
        self.control = control
        self.parameters = parameters

        if host is None:
            # AcquisitionService is imported only on the Red Pitaya since pyrp3 is not
            # available on Windows
            from linien_server.acquisition import AcquisitionService

            self.acquisition = AcquisitionService()
        else:
            # AcquisitionService has to be started manually on the Red Pitaya
            self.acquisition = rpyc.connect(host, ACQUISITION_PORT).root

        self._last_sweep_speed = None
        self._last_raw_acquisition_settings = None
        self._iir_cache = {}  # type: ignore[var-annotated]

        self.parameters.lock.add_callback(self.acquisition.exposed_set_lock_status)
        self.parameters.fetch_additional_signals.add_callback(
            self.acquisition.exposed_set_fetch_additional_signals  # noqa: E501
        )
        self.parameters.dual_channel.add_callback(
            self.acquisition.exposed_set_dual_channel
        )

    def write_registers(self):
        """Writes data from `parameters` to the FPGA."""

        max_ = lambda val: val if np.abs(val) <= 8191 else (8191 * val / np.abs(val))
        phase_to_delay = lambda phase: int(phase / 360 * (1 << 14))

        if not self.parameters.dual_channel.value:
            factor_a = 256
            factor_b = 0
        else:
            factor_a, factor_b = convert_channel_mixing_value(
                self.parameters.channel_mixing.value
            )

        lock_changed = self.parameters.lock.value != self.control.exposed_is_locked
        self.control.exposed_is_locked = self.parameters.lock.value

        new = dict(
            # sweep run is 1 by default. The gateware automatically takes care of
            # stopping the sweep run after `request_lock` is set by setting
            # `sweep.clear`
            logic_sweep_run=1,
            logic_sweep_pause=int(self.parameters.sweep_pause.value),
            logic_sweep_step=int(
                DEFAULT_SWEEP_SPEED
                * self.parameters.sweep_amplitude.value
                / (2**self.parameters.sweep_speed.value)
            ),
            # NOTE: Sweep center is set by `logic_out_offset`.
            logic_sweep_min=-1 * max_(self.parameters.sweep_amplitude.value * 8191),
            logic_sweep_max=max_(self.parameters.sweep_amplitude.value * 8191),
            logic_mod_freq=self.parameters.modulation_frequency.value,
            logic_mod_amp=self.parameters.modulation_amplitude.value
            if self.parameters.modulation_frequency.value > 0
            else 0,
            logic_dual_channel=int(self.parameters.dual_channel.value),
            logic_fast_mode=int(self.parameters.fast_mode.value),
            logic_chain_a_factor=factor_a,
            logic_chain_b_factor=factor_b,
            logic_chain_a_offset=twos_complement(
                int(self.parameters.offset_a.value), 14
            ),
            logic_chain_b_offset=twos_complement(
                int(self.parameters.offset_b.value), 14
            ),
            logic_out_offset=int(self.parameters.sweep_center.value * 8191),
            logic_combined_offset=twos_complement(
                self.parameters.combined_offset.value, 14
            ),
            logic_control_channel=self.parameters.control_channel.value,
            logic_mod_channel=self.parameters.mod_channel.value,
            logic_sweep_channel=self.parameters.sweep_channel.value,
            logic_slow_control_channel=self.parameters.slow_control_channel.value,
            slow_chain_pid_reset=not self.parameters.pid_on_slow_enabled.value,
            logic_analog_out_1=self.parameters.analog_out_1.value,
            logic_analog_out_2=self.parameters.analog_out_2.value,
            logic_analog_out_3=self.parameters.analog_out_3.value,
            logic_autolock_fast_target_position=self.parameters.autolock_target_position.value,  # noqa: E501
            logic_autolock_autolock_mode=self.parameters.autolock_mode.value,
            logic_autolock_robust_N_instructions=len(
                self.parameters.autolock_instructions.value
            ),
            logic_autolock_robust_time_scale=self.parameters.autolock_time_scale.value,
            logic_autolock_robust_final_wait_time=self.parameters.autolock_final_wait_time.value,  # noqa: E501
            # channel A
            fast_a_demod_delay=phase_to_delay(
                self.parameters.demodulation_phase_a.value
            )
            if self.parameters.modulation_frequency.value > 0
            else 0,
            fast_a_demod_multiplier=self.parameters.demodulation_multiplier_a.value,
            fast_a_dx_sel=csrmap.signals.index("zero"),
            fast_a_y_tap=2,
            fast_a_dy_sel=csrmap.signals.index("zero"),
            fast_a_invert=int(self.parameters.invert_a.value),
            # channel B
            fast_b_demod_delay=phase_to_delay(
                self.parameters.demodulation_phase_b.value
            )
            if self.parameters.modulation_frequency.value > 0
            else 0,
            fast_b_demod_multiplier=self.parameters.demodulation_multiplier_b.value,
            fast_b_dx_sel=csrmap.signals.index("zero"),
            fast_b_y_tap=1,
            fast_b_dy_sel=csrmap.signals.index("zero"),
            fast_b_invert=int(self.parameters.invert_b.value),
            # trigger on sweep
            scopegen_external_trigger=1,
            gpio_p_oes=0b11111111,
            gpio_n_oes=0b11111111,
            gpio_p_outs=self.parameters.gpio_p_out.value,
            gpio_n_outs=self.parameters.gpio_n_out.value,
            gpio_n_do0_en=csrmap.signals.index("zero"),
            gpio_n_do1_en=csrmap.signals.index("zero"),
            logic_slow_decimation=16,
        )

        for instruction_idx, [wait_for, peak_height] in enumerate(
            self.parameters.autolock_instructions.value
        ):
            new["logic_autolock_robust_peak_height_%d" % instruction_idx] = peak_height
            new["logic_autolock_robust_wait_for_%d" % instruction_idx] = wait_for

        if self.parameters.lock.value:
            # display combined error signal and control signal
            new.update(
                {
                    "scopegen_adc_a_sel": csrmap.signals.index(
                        "logic_combined_error_signal"
                        if not self.parameters.acquisition_raw_filter_enabled.value
                        else "logic_combined_error_signal_filtered"
                    ),
                    "scopegen_adc_a_q_sel": csrmap.signals.index("fast_b_x"),
                    "scopegen_adc_b_sel": csrmap.signals.index("logic_control_signal"),
                    "scopegen_adc_b_q_sel": csrmap.signals.index("zero"),
                }
            )
        else:
            # display both demodulated error signals (if dual channel mode) OR: display
            # demodulated error signal 1 + monitor signal
            new.update(
                {
                    "scopegen_adc_a_sel": csrmap.signals.index("fast_a_out_i"),
                    "scopegen_adc_a_q_sel": csrmap.signals.index("fast_a_out_q"),
                    "scopegen_adc_b_sel": csrmap.signals.index(
                        "fast_b_out_i"
                        if self.parameters.dual_channel.value
                        else "fast_b_x"
                    ),
                    "scopegen_adc_b_q_sel": csrmap.signals.index(
                        "fast_b_out_q" if self.parameters.dual_channel.value else "zero"
                    ),
                }
            )

        # filter out values that did not change
        new = dict(
            (k, v)
            for k, v in new.items()
            if (
                (k not in self.control._cached_data)
                or (self.control._cached_data.get(k) != v)
            )
        )
        self.control._cached_data.update(new)

        # pass sweep speed changes to acquisition process
        sweep_changed = self.parameters.sweep_speed.value != self._last_sweep_speed
        if sweep_changed:
            self._last_sweep_speed = self.parameters.sweep_speed.value
            self.acquisition.exposed_set_sweep_speed(self.parameters.sweep_speed.value)

        raw_acquisition_settings = (
            self.parameters.acquisition_raw_enabled.value,
            self.parameters.acquisition_raw_decimation.value,
        )
        if raw_acquisition_settings != self._last_raw_acquisition_settings:
            self._last_raw_acquisition_settings = raw_acquisition_settings
            self.acquisition.exposed_set_raw_acquisition(*raw_acquisition_settings)

        fpga_base_freq = 125e6

        self.set_iir(
            "logic_raw_acquisition_iir",
            *make_filter(
                "LP",
                f=self.parameters.acquisition_raw_filter_frequency.value
                / fpga_base_freq,
                k=1,
            ),
        )

        for k, v in new.items():
            self.set(k, int(v))

        if not self.parameters.lock.value and sweep_changed:
            # reset sweep for a short time if the scan range was changed this is needed
            # because otherwise it may take too long before the new scan range is
            # reached --> no scope trigger is sent
            self.set("logic_sweep_run", 0)
            self.set("logic_sweep_run", 1)

        kp = self.parameters.p.value
        ki = self.parameters.i.value
        kd = self.parameters.d.value
        slope = self.parameters.target_slope_rising.value
        control_channel, sweep_channel, slow_control_channel = (
            self.parameters.control_channel.value,
            self.parameters.sweep_channel.value,
            self.parameters.slow_control_channel.value,
        )

        def channel_polarity(channel):
            return (
                self.parameters.polarity_fast_out1.value,
                self.parameters.polarity_fast_out2.value,
                self.parameters.polarity_analog_out0.value,
            )[channel]

        if control_channel != sweep_channel:
            if channel_polarity(control_channel) != channel_polarity(sweep_channel):
                slope = not slope

        slow_strength = (
            self.parameters.pid_on_slow_strength.value
            if self.parameters.pid_on_slow_enabled.value
            else 0
        )
        slow_slope = (
            1
            if channel_polarity(slow_control_channel)
            == channel_polarity(control_channel)
            else -1
        )

        for chain in ("a", "b"):
            automatic = getattr(self.parameters, "filter_automatic_%s" % chain).value
            # iir_idx means iir_c or iir_d
            for iir_idx in range(2):
                # iir_sub_idx means in-phase signal or quadrature signal
                for iir_sub_idx in range(2):
                    iir_name = "fast_%s_iir_%s_%d" % (
                        chain,
                        ("c", "d")[iir_idx],
                        iir_sub_idx + 1,
                    )

                    if automatic:
                        filter_enabled = True
                        filter_type = LOW_PASS_FILTER
                        filter_frequency = (
                            self.parameters.modulation_frequency.value / MHz * 1e6 / 2
                        )

                        # if the filter frequency is too low (< 10Hz), the IIR doesn't
                        # work properly anymore. In that case, don't filter. This is
                        # also helpful if the raw (not demodulated) signal should be
                        # displayed which can be achieved by setting modulation
                        # frequency to 0.
                        if filter_frequency < 10:
                            filter_enabled = False
                    else:
                        filter_enabled = getattr(
                            self.parameters,
                            "filter_%d_enabled_%s" % (iir_idx + 1, chain),
                        ).value
                        filter_type = getattr(
                            self.parameters, "filter_%d_type_%s" % (iir_idx + 1, chain)
                        ).value
                        filter_frequency = getattr(
                            self.parameters,
                            "filter_%d_frequency_%s" % (iir_idx + 1, chain),
                        ).value

                    if not filter_enabled:
                        self.set_iir(iir_name, *make_filter("P", k=1))
                    else:
                        if filter_type == LOW_PASS_FILTER:
                            self.set_iir(
                                iir_name,
                                *make_filter(
                                    "LP", f=filter_frequency / fpga_base_freq, k=1
                                ),
                            )
                        elif filter_type == HIGH_PASS_FILTER:
                            self.set_iir(
                                iir_name,
                                *make_filter(
                                    "HP", f=filter_frequency / fpga_base_freq, k=1
                                ),
                            )
                        else:
                            raise Exception(
                                "unknown filter %s for %s" % (filter_type, iir_name)
                            )

        if lock_changed:
            if self.parameters.lock.value:
                # set PI parameters
                self.set_pid(kp, ki, kd, slope, reset=0, request_lock=1)
                self.set_slow_pid(slow_strength, slow_slope, reset=0)
            else:
                self.set_pid(0, 0, 0, slope, reset=1, request_lock=0)
                self.set_slow_pid(0, slow_slope, reset=1)
        else:
            if self.parameters.lock.value:
                # set new PI parameters
                self.set_pid(kp, ki, kd, slope)
                self.set_slow_pid(slow_strength, slow_slope)

    def set_pid(self, p, i, d, slope, reset=None, request_lock=None):
        if request_lock is not None:
            self.set("logic_autolock_request_lock", request_lock)

        sign = -1 if slope else 1
        self.set("logic_pid_kp", p * sign)
        self.set("logic_pid_ki", i * sign)
        self.set("logic_pid_kd", d * sign)

        if reset is not None:
            self.set("logic_pid_reset", reset)

    def set_slow_pid(self, strength, slope, reset=None):
        sign = slope
        self.set("slow_chain_pid_ki", strength * sign)

        if reset is not None:
            self.set("slow_chain_pid_reset", reset)

    def set(self, key, value):
        self.acquisition.exposed_set_csr(key, value)

    def set_iir(self, iir_name, *args):
        if self._iir_cache.get(iir_name) != args:
            # as setting iir parameters takes some time, take care that we don't  do it
            # too often
            self.acquisition.exposed_set_iir_csr(iir_name, *args)
            self._iir_cache[iir_name] = args


def twos_complement(num, N_bits):
    max_ = 1 << (N_bits - 1)
    full = 2 * max_
    if num < 0:
        num += full
    return num



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from time import time

import numpy as np
from linien_common.common import determine_shift_by_correlation

ZOOM_STEP = 2


class Approacher:
    def __init__(
        self,
        control,
        parameters,
        first_error_signal,
        target_zoom,
        central_y,
        allow_sweep_speed_change=False,
        wait_time_between_current_corrections=None,
    ):
        self.control = control
        self.parameters = parameters
        # central_y is the y coordinate between maximum and minimum of the
        # target line. We vertically center the target line with respect to the
        # x axis because correlation doesn't work if the signal is only positive
        self.first_error_signal = first_error_signal - central_y
        self.target_zoom = target_zoom
        self.allow_sweep_speed_change = allow_sweep_speed_change

        self.wait_time_between_current_corrections = (
            wait_time_between_current_corrections
        )

        self.central_y = central_y

        self.reset_properties()

    def reset_properties(self):
        self.history = []
        self.zoom_factor = 1
        self.N_at_this_zoom = 0
        self.last_shifts_at_this_zoom = None
        self.time_last_current_correction = None
        self.time_last_zoom = time()

    def approach_line(self, error_signal):
        if time() - self.time_last_zoom > 15:
            raise Exception("approaching took too long")

        error_signal = error_signal - self.central_y

        initial_sweep_amplitude = self.parameters.autolock_initial_sweep_amplitude.value

        # the autolock tries to center a line by changing the sweep center.
        # If a line was selected that is close to the edges, this can lead to
        # a situation where sweep center + sweep_amplitude > output limits of RP.
        # in this case, we want to ignore the error signal that was recorded at
        # these points as it may contain a distorted version of the spectrum that
        # disturbs the correlation.
        sweep_amplitude = self.parameters.sweep_amplitude.value
        center = self.parameters.sweep_center.value
        sweep = (
            np.linspace(-sweep_amplitude, sweep_amplitude, len(error_signal)) + center
        )
        error_signal = np.array(error_signal)
        error_signal[np.abs(sweep) > 1] = np.nan

        # now, we calculate the correlation to find the shift
        shift, zoomed_ref, zoomed_err = determine_shift_by_correlation(
            self.zoom_factor, self.first_error_signal, error_signal
        )
        shift *= initial_sweep_amplitude
        self.history.append((zoomed_ref, zoomed_err))
        self.history.append("shift %f" % (-1 * shift))

        if self.N_at_this_zoom == 0:
            # if we are at the final zoom, we should be very quick.
            # Therefore, we just correct the current and turn the lock on
            # immediately. We skip the rest of this method (drift detection etc.)
            next_step_is_lock = self.zoom_factor >= self.target_zoom
            if next_step_is_lock:
                return True
            else:
                self._correct_current(shift)
        else:
            # wait for some time after the last current correction
            min_wait_time = (
                1
                if self.wait_time_between_current_corrections is None
                else self.wait_time_between_current_corrections
            )
            if time() - self.time_last_current_correction < min_wait_time:
                return

            # check that the drift is slow
            # this is needed for systems that only react slowly to changes in
            # input parameters. In this case, we have to wait until the reaction
            # to the last input is done.
            shift_diff = np.abs(shift - self.last_shifts_at_this_zoom[-1])
            drift_slow = shift_diff < initial_sweep_amplitude / self.target_zoom / 8

            # if data comes in very slowly (<1 Hz), we skip the drift analysis
            # because it would take too much time
            low_recording_rate = self.parameters.sweep_speed.value > 10

            if low_recording_rate or drift_slow:
                is_close_to_target = shift < self.parameters.sweep_amplitude.value / 8
                if is_close_to_target:
                    return self._decrease_scan_range()
                else:
                    self._correct_current(shift)

        self.N_at_this_zoom += 1
        self.last_shifts_at_this_zoom = self.last_shifts_at_this_zoom or []
        self.last_shifts_at_this_zoom.append(shift)

    def _decrease_scan_range(self):
        self.N_at_this_zoom = 0
        self.last_shifts_at_this_zoom = None

        self.zoom_factor *= ZOOM_STEP
        self.time_last_zoom = time()

        self.control.exposed_pause_acquisition()

        self.parameters.sweep_amplitude.value /= ZOOM_STEP
        if self.allow_sweep_speed_change:
            new_sweep_speed = (
                self.parameters.sweep_speed.value - 1
                if self.parameters.sweep_speed.value > 5
                else self.parameters.sweep_speed.value
            )
            self.parameters.sweep_speed.value = new_sweep_speed
        self.control.exposed_write_registers()
        self.control.exposed_continue_acquisition()

    def _correct_current(self, shift):
        self.control.exposed_pause_acquisition()
        self.time_last_current_correction = time()

        self.parameters.sweep_center.value -= shift

        self.control.exposed_write_registers()
        self.control.exposed_continue_acquisition()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.signal import correlate


def get_lock_region(spectrum, target_idxs):
    """Given a spectrum and the points that the user selected for locking,
    calculate the region where locking will work. This is the region between the
    next zero crossings after the extrema."""
    part = spectrum[target_idxs[0] : target_idxs[1]]
    extrema = tuple(
        sorted([target_idxs[0] + np.argmin(part), target_idxs[0] + np.argmax(part)])
    )

    def walk_until_sign_changes(start_idx, direction):
        current_idx = start_idx
        start_sign = sign(spectrum[start_idx])
        while True:
            current_idx += direction
            if current_idx < 0:
                return 0
            if current_idx == len(spectrum):
                return current_idx - 1

            current_value = sign(spectrum[current_idx])
            current_sign = sign(current_value)

            if current_sign != start_sign:
                return current_idx - direction

    return (
        walk_until_sign_changes(extrema[0], -1),
        walk_until_sign_changes(extrema[1], 1),
    )


def get_time_scale(spectrum, target_idxs):
    part = spectrum[target_idxs[0] : target_idxs[1]]
    return np.abs(np.argmin(part) - np.argmax(part))


def sum_up_spectrum(spectrum):
    sum_ = 0
    summed = []

    for value in spectrum:
        summed.append(sum_ + value)
        sum_ += value

    return summed


def get_diff_at_time_scale(summed, xscale):
    new = []

    for idx, value in enumerate(summed):
        if idx < xscale:
            old = 0
        else:
            old = summed[idx - xscale]

        new.append(value - old)

    return new


def sign(value):
    return 1 if value >= 1 else -1


def get_target_peak(summed_xscaled, target_idxs):
    selected_region = summed_xscaled[target_idxs[0] : target_idxs[1]]
    # in the selected region, we may have 1 minimum and one maximum
    # we know that we are interested in the "left" extremum --> sort extrema
    # by index and take the first one
    extremum = np.min([np.argmin(selected_region), np.argmax(selected_region)])
    current_idx = target_idxs[0] + extremum
    return current_idx


def get_all_peaks(summed_xscaled, target_idxs):
    current_idx = get_target_peak(summed_xscaled, target_idxs)

    peaks = []

    peaks.append((current_idx, summed_xscaled[current_idx]))

    while True:
        if current_idx == 0:
            break
        current_idx -= 1

        value = summed_xscaled[current_idx]
        last_peak_position, last_peak_height = peaks[-1]

        if sign(last_peak_height) == sign(value):
            if np.abs(value) > np.abs(last_peak_height):
                peaks[-1] = (current_idx, value)
        else:
            peaks.append((current_idx, value))

    return peaks


def crop_spectra_to_same_view(spectra_with_jitter):
    cropped_spectra = []

    shifts = []

    for idx, spectrum in enumerate(spectra_with_jitter):
        shift = np.argmax(correlate(spectra_with_jitter[0], spectrum)) - len(spectrum)

        shifts.append(-shift)

    min_shift = min(shifts)
    max_shift = max(shifts)

    length_after_crop = len(spectra_with_jitter[0]) - (max_shift - min_shift)

    for shift, spectrum in zip(shifts, spectra_with_jitter):
        cropped_spectra.append(spectrum[shift - min_shift :][:length_after_crop])

    return cropped_spectra, -min_shift + 1



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import (
    AUTO_DETECT_AUTOLOCK_MODE,
    FAST_AUTOLOCK,
    N_POINTS,
    ROBUST_AUTOLOCK,
    determine_shift_by_correlation,
)


class AutolockAlgorithmSelector:
    """This class helps deciding which autolock method should be used."""

    def __init__(
        self,
        mode_preference,
        spectrum,
        additional_spectra,
        line_width,
        N_spectra_required=3,
    ):
        self.done = False
        self.mode = None
        self.spectra = [spectrum] + (additional_spectra or [])
        self.N_spectra_required = N_spectra_required
        self.line_width = line_width

        if mode_preference != AUTO_DETECT_AUTOLOCK_MODE:
            self.mode = mode_preference
            self.done = True
            return

        self.check()

    def handle_new_spectrum(self, spectrum):
        self.spectra.append(spectrum)
        self.check()

    def check(self):
        if self.done:
            return True

        if len(self.spectra) < self.N_spectra_required:
            return
        else:
            ref = self.spectra[0]
            additional = self.spectra[1:]
            abs_shifts = [
                abs(determine_shift_by_correlation(1, ref, spectrum)[0] * N_POINTS)
                for spectrum in additional
            ]
            max_shift = max(abs_shifts)
            print("jitter / line width ratio:", max_shift / (self.line_width / 2))

            if max_shift <= self.line_width / 2:
                self.mode = FAST_AUTOLOCK
            else:
                self.mode = ROBUST_AUTOLOCK

            self.done = True



# Copyright 2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from threading import Event, Thread
from time import sleep
from typing import Tuple

import requests
from linien_common.influxdb import InfluxDBCredentials, save_credentials
from linien_server.parameters import Parameters


class InfluxDBLogger:
    def __init__(
        self, credentials: InfluxDBCredentials, parameters: Parameters
    ) -> None:
        self.credentials = credentials
        self.parameters = parameters
        self.stop_event = Event()
        self.stop_event.set()

    @property
    def credentials(self) -> InfluxDBCredentials:
        return self._credentials

    @credentials.setter
    def credentials(self, value: InfluxDBCredentials) -> None:
        self._credentials = value
        save_credentials(value)

    def start_logging(self, interval: float) -> None:
        conn_success, status_code, message = self.test_connection(self.credentials)
        if conn_success:
            self.stop_event.clear()
            self.thread = Thread(
                target=self._logging_loop,
                args=(interval,),
                daemon=True,
            )
            self.thread.start()
        else:
            raise ConnectionError(
                "Failed to connect to InfluxDB database: %s (Status code: %s)"
                % (message, status_code)
            )

    def stop_logging(self) -> None:
        self.stop_event.set()
        self.thread.join()

    def _logging_loop(self, interval: float) -> None:
        while not self.stop_event.is_set():
            data = {}
            for name, param in self.parameters:
                if param.log:
                    if name == "signal_stats":
                        for stat_name, stat_value in param.value.items():
                            data[stat_name] = stat_value
                    else:
                        data[name] = param.value
            self.write_data(self.credentials, data)
            sleep(interval)

    def test_connection(
        self, credentials: InfluxDBCredentials
    ) -> Tuple[bool, int, str]:
        """Write empty data to the server to test the connection"""
        response = self.write_data(credentials, data={})
        success = response.status_code == 204
        return success, response.status_code, response.text

    def write_data(
        self, credentials: InfluxDBCredentials, data: dict
    ) -> requests.Response:
        """Write data to the database"""
        endpoint = credentials.url + "/api/v2/write"
        headers = {
            "Authorization": "Token " + credentials.token,
            "Content-Type": "text/plain; charset=utf-8",
            "Accept": "application/json",
        }
        params = {
            "org": credentials.org,
            "bucket": credentials.bucket,
            "precision": "ns",
        }

        point = self._convert_to_line_protocol(data)

        response = requests.post(endpoint, headers=headers, params=params, data=point)
        return response

    def _convert_to_line_protocol(self, data: dict) -> str:
        if not data:
            return ""
        point = self.credentials.measurement
        for i, (key, value) in enumerate(data.items()):
            if i == 0:
                point += " "
            else:
                point += ","
            point += "%s=%s" % (key, value)
        return point



# Copyright 2014-2015 Robert Jördens <jordens@gmail.com>
# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.


from . import csrmap
from .iir_coeffs import get_params


class PythonCSR:
    map = csrmap.csr
    constants = csrmap.csr_constants
    offset = 0x40300000

    def __init__(self, rp):
        self.rp = rp

    def set_one(self, addr: str, value: int) -> None:
        self.rp.write(addr, value)

    def get_one(self, addr):
        return int(self.rp.read(addr))

    def set(self, name, value):
        map, addr, width, wr = self.map[name]
        assert wr, name

        ma = 1 << width
        bit_mask = ma - 1
        val = value & bit_mask
        assert value == val or ma + value == val, (
            "value for %s out of range" % name,
            (value, val, ma),
        )

        b = (width + 8 - 1) // 8
        for i in range(b):
            v = (val >> (8 * (b - i - 1))) & 0xFF
            self.set_one(self.offset + (map << 11) + ((addr + i) << 2), v)

    def get(self, name: str) -> int:
        if name in self.constants:
            return self.constants[name]

        map, addr, nr, wr = self.map[name]
        v = 0
        b = (nr + 8 - 1) // 8
        for i in range(b):
            v |= self.get_one(self.offset + (map << 11) + ((addr + i) << 2)) << 8 * (
                b - i - 1
            )
        return v

    def set_iir(self, prefix, b, a, z=0):
        shift = self.get(prefix + "_shift") or 16
        width = self.get(prefix + "_width") or 18
        interval = self.get(prefix + "_interval") or 1
        b, a, params = get_params(b, a, shift, width, interval)

        for k in sorted(params):
            self.set(prefix + "_" + k, params[k])
        self.set(prefix + "_z0", z)
        for i in range(len(b), 3):
            n = prefix + "_b%i" % i
            if n in self.map:
                self.set(n, 0)
                self.set(prefix + "_a%i" % i, 0)

    def states(self, *names):
        return sum(1 << csrmap.states.index(name) for name in names)



csr_constants = {
    'fast_a_iir_c_1_shift': 23,
    'fast_a_iir_c_1_width': 25,
    'fast_a_iir_c_1_interval': 1,
    'fast_a_iir_c_1_latency': 2,
    'fast_a_iir_c_1_order': 1,
    'fast_a_iir_c_1_iterative': 0,
    'fast_a_iir_d_1_shift': 23,
    'fast_a_iir_d_1_width': 25,
    'fast_a_iir_d_1_interval': 1,
    'fast_a_iir_d_1_latency': 3,
    'fast_a_iir_d_1_order': 2,
    'fast_a_iir_d_1_iterative': 0,
    'fast_a_iir_c_2_shift': 23,
    'fast_a_iir_c_2_width': 25,
    'fast_a_iir_c_2_interval': 1,
    'fast_a_iir_c_2_latency': 2,
    'fast_a_iir_c_2_order': 1,
    'fast_a_iir_c_2_iterative': 0,
    'fast_a_iir_d_2_shift': 23,
    'fast_a_iir_d_2_width': 25,
    'fast_a_iir_d_2_interval': 1,
    'fast_a_iir_d_2_latency': 3,
    'fast_a_iir_d_2_order': 2,
    'fast_a_iir_d_2_iterative': 0,
    'fast_b_iir_c_1_shift': 23,
    'fast_b_iir_c_1_width': 25,
    'fast_b_iir_c_1_interval': 1,
    'fast_b_iir_c_1_latency': 2,
    'fast_b_iir_c_1_order': 1,
    'fast_b_iir_c_1_iterative': 0,
    'fast_b_iir_d_1_shift': 23,
    'fast_b_iir_d_1_width': 25,
    'fast_b_iir_d_1_interval': 1,
    'fast_b_iir_d_1_latency': 3,
    'fast_b_iir_d_1_order': 2,
    'fast_b_iir_d_1_iterative': 0,
    'fast_b_iir_c_2_shift': 23,
    'fast_b_iir_c_2_width': 25,
    'fast_b_iir_c_2_interval': 1,
    'fast_b_iir_c_2_latency': 2,
    'fast_b_iir_c_2_order': 1,
    'fast_b_iir_c_2_iterative': 0,
    'fast_b_iir_d_2_shift': 23,
    'fast_b_iir_d_2_width': 25,
    'fast_b_iir_d_2_interval': 1,
    'fast_b_iir_d_2_latency': 3,
    'fast_b_iir_d_2_order': 2,
    'fast_b_iir_d_2_iterative': 0,
    'logic_raw_acquisition_iir_shift': 23,
    'logic_raw_acquisition_iir_width': 25,
    'logic_raw_acquisition_iir_interval': 1,
    'logic_raw_acquisition_iir_latency': 6,
    'logic_raw_acquisition_iir_order': 5,
    'logic_raw_acquisition_iir_iterative': 0,
}

csr = {
    'dna_dna': (28, 0x000, 64, False),
    'fast_a_y_tap': (0, 0x000, 2, True),
    'fast_a_invert': (0, 0x001, 1, True),
    'fast_a_demod_delay': (0, 0x002, 32, True),
    'fast_a_demod_multiplier': (0, 0x006, 4, True),
    'fast_a_x_limit_1_min': (0, 0x007, 25, True),
    'fast_a_x_limit_1_max': (0, 0x00b, 25, True),
    'fast_a_iir_c_1_z0': (0, 0x00f, 27, True),
    'fast_a_iir_c_1_a1': (0, 0x013, 25, True),
    'fast_a_iir_c_1_b0': (0, 0x017, 25, True),
    'fast_a_iir_c_1_b1': (0, 0x01b, 25, True),
    'fast_a_iir_d_1_z0': (0, 0x01f, 27, True),
    'fast_a_iir_d_1_a1': (0, 0x023, 25, True),
    'fast_a_iir_d_1_a2': (0, 0x027, 25, True),
    'fast_a_iir_d_1_b0': (0, 0x02b, 25, True),
    'fast_a_iir_d_1_b1': (0, 0x02f, 25, True),
    'fast_a_iir_d_1_b2': (0, 0x033, 25, True),
    'fast_a_y_limit_1_min': (0, 0x037, 25, True),
    'fast_a_y_limit_1_max': (0, 0x03b, 25, True),
    'fast_a_x_limit_2_min': (0, 0x03f, 25, True),
    'fast_a_x_limit_2_max': (0, 0x043, 25, True),
    'fast_a_iir_c_2_z0': (0, 0x047, 27, True),
    'fast_a_iir_c_2_a1': (0, 0x04b, 25, True),
    'fast_a_iir_c_2_b0': (0, 0x04f, 25, True),
    'fast_a_iir_c_2_b1': (0, 0x053, 25, True),
    'fast_a_iir_d_2_z0': (0, 0x057, 27, True),
    'fast_a_iir_d_2_a1': (0, 0x05b, 25, True),
    'fast_a_iir_d_2_a2': (0, 0x05f, 25, True),
    'fast_a_iir_d_2_b0': (0, 0x063, 25, True),
    'fast_a_iir_d_2_b1': (0, 0x067, 25, True),
    'fast_a_iir_d_2_b2': (0, 0x06b, 25, True),
    'fast_a_y_limit_2_min': (0, 0x06f, 25, True),
    'fast_a_y_limit_2_max': (0, 0x073, 25, True),
    'fast_a_x_clr': (0, 0x077, 1, True),
    'fast_a_x_max': (0, 0x078, 25, False),
    'fast_a_x_min': (0, 0x07c, 25, False),
    'fast_a_out_i_clr': (0, 0x080, 1, True),
    'fast_a_out_i_max': (0, 0x081, 25, False),
    'fast_a_out_i_min': (0, 0x085, 25, False),
    'fast_a_out_q_clr': (0, 0x089, 1, True),
    'fast_a_out_q_max': (0, 0x08a, 25, False),
    'fast_a_out_q_min': (0, 0x08e, 25, False),
    'fast_a_dx_sel': (0, 0x092, 4, True),
    'fast_a_dy_sel': (0, 0x093, 4, True),
    'fast_b_y_tap': (1, 0x000, 2, True),
    'fast_b_invert': (1, 0x001, 1, True),
    'fast_b_demod_delay': (1, 0x002, 32, True),
    'fast_b_demod_multiplier': (1, 0x006, 4, True),
    'fast_b_x_limit_1_min': (1, 0x007, 25, True),
    'fast_b_x_limit_1_max': (1, 0x00b, 25, True),
    'fast_b_iir_c_1_z0': (1, 0x00f, 27, True),
    'fast_b_iir_c_1_a1': (1, 0x013, 25, True),
    'fast_b_iir_c_1_b0': (1, 0x017, 25, True),
    'fast_b_iir_c_1_b1': (1, 0x01b, 25, True),
    'fast_b_iir_d_1_z0': (1, 0x01f, 27, True),
    'fast_b_iir_d_1_a1': (1, 0x023, 25, True),
    'fast_b_iir_d_1_a2': (1, 0x027, 25, True),
    'fast_b_iir_d_1_b0': (1, 0x02b, 25, True),
    'fast_b_iir_d_1_b1': (1, 0x02f, 25, True),
    'fast_b_iir_d_1_b2': (1, 0x033, 25, True),
    'fast_b_y_limit_1_min': (1, 0x037, 25, True),
    'fast_b_y_limit_1_max': (1, 0x03b, 25, True),
    'fast_b_x_limit_2_min': (1, 0x03f, 25, True),
    'fast_b_x_limit_2_max': (1, 0x043, 25, True),
    'fast_b_iir_c_2_z0': (1, 0x047, 27, True),
    'fast_b_iir_c_2_a1': (1, 0x04b, 25, True),
    'fast_b_iir_c_2_b0': (1, 0x04f, 25, True),
    'fast_b_iir_c_2_b1': (1, 0x053, 25, True),
    'fast_b_iir_d_2_z0': (1, 0x057, 27, True),
    'fast_b_iir_d_2_a1': (1, 0x05b, 25, True),
    'fast_b_iir_d_2_a2': (1, 0x05f, 25, True),
    'fast_b_iir_d_2_b0': (1, 0x063, 25, True),
    'fast_b_iir_d_2_b1': (1, 0x067, 25, True),
    'fast_b_iir_d_2_b2': (1, 0x06b, 25, True),
    'fast_b_y_limit_2_min': (1, 0x06f, 25, True),
    'fast_b_y_limit_2_max': (1, 0x073, 25, True),
    'fast_b_x_clr': (1, 0x077, 1, True),
    'fast_b_x_max': (1, 0x078, 25, False),
    'fast_b_x_min': (1, 0x07c, 25, False),
    'fast_b_out_i_clr': (1, 0x080, 1, True),
    'fast_b_out_i_max': (1, 0x081, 25, False),
    'fast_b_out_i_min': (1, 0x085, 25, False),
    'fast_b_out_q_clr': (1, 0x089, 1, True),
    'fast_b_out_q_max': (1, 0x08a, 25, False),
    'fast_b_out_q_min': (1, 0x08e, 25, False),
    'fast_b_dx_sel': (1, 0x092, 4, True),
    'fast_b_dy_sel': (1, 0x093, 4, True),
    'gpio_n_ins': (30, 0x000, 8, False),
    'gpio_n_outs': (30, 0x001, 8, True),
    'gpio_n_oes': (30, 0x002, 8, True),
    'gpio_n_state': (30, 0x003, 14, False),
    'gpio_n_state_clr': (30, 0x005, 1, True),
    'gpio_n_do0_en': (30, 0x006, 14, True),
    'gpio_n_do1_en': (30, 0x008, 14, True),
    'gpio_n_do2_en': (30, 0x00a, 14, True),
    'gpio_n_do3_en': (30, 0x00c, 14, True),
    'gpio_n_do4_en': (30, 0x00e, 14, True),
    'gpio_n_do5_en': (30, 0x010, 14, True),
    'gpio_n_do6_en': (30, 0x012, 14, True),
    'gpio_n_do7_en': (30, 0x014, 14, True),
    'gpio_p_ins': (31, 0x000, 8, False),
    'gpio_p_outs': (31, 0x001, 8, True),
    'gpio_p_oes': (31, 0x002, 8, True),
    'logic_dual_channel': (8, 0x000, 1, True),
    'logic_mod_channel': (8, 0x001, 1, True),
    'logic_control_channel': (8, 0x002, 1, True),
    'logic_sweep_channel': (8, 0x003, 2, True),
    'logic_slow_control_channel': (8, 0x004, 2, True),
    'logic_fast_mode': (8, 0x005, 1, True),
    'logic_chain_a_factor': (8, 0x006, 9, True),
    'logic_chain_b_factor': (8, 0x008, 9, True),
    'logic_chain_a_offset': (8, 0x00a, 14, True),
    'logic_chain_b_offset': (8, 0x00c, 14, True),
    'logic_combined_offset': (8, 0x00e, 14, True),
    'logic_out_offset': (8, 0x010, 14, True),
    'logic_slow_decimation': (8, 0x012, 5, True),
    'logic_analog_out_1': (8, 0x013, 15, True),
    'logic_analog_out_2': (8, 0x015, 15, True),
    'logic_analog_out_3': (8, 0x017, 15, True),
    'logic_slow_value': (8, 0x019, 14, False),
    'logic_mod_amp': (8, 0x01b, 14, True),
    'logic_mod_freq': (8, 0x01d, 32, True),
    'logic_sweep_step': (8, 0x021, 30, True),
    'logic_sweep_min': (8, 0x025, 14, True),
    'logic_sweep_max': (8, 0x027, 14, True),
    'logic_sweep_run': (8, 0x029, 1, True),
    'logic_sweep_pause': (8, 0x02a, 1, True),
    'logic_limit_error_signal_min': (8, 0x02b, 25, True),
    'logic_limit_error_signal_max': (8, 0x02f, 25, True),
    'logic_limit_fast1_min': (8, 0x033, 14, True),
    'logic_limit_fast1_max': (8, 0x035, 14, True),
    'logic_limit_fast2_min': (8, 0x037, 14, True),
    'logic_limit_fast2_max': (8, 0x039, 14, True),
    'logic_pid_setpoint': (8, 0x03b, 25, True),
    'logic_pid_kp': (8, 0x03f, 14, True),
    'logic_pid_ki': (8, 0x041, 14, True),
    'logic_pid_reset': (8, 0x043, 1, True),
    'logic_pid_kd': (8, 0x044, 14, True),
    'logic_autolock_robust_time_scale': (8, 0x046, 14, True),
    'logic_autolock_robust_N_instructions': (8, 0x048, 5, True),
    'logic_autolock_robust_final_wait_time': (8, 0x049, 14, True),
    'logic_autolock_robust_peak_height_0': (8, 0x04b, 28, True),
    'logic_autolock_robust_peak_height_1': (8, 0x04f, 28, True),
    'logic_autolock_robust_peak_height_2': (8, 0x053, 28, True),
    'logic_autolock_robust_peak_height_3': (8, 0x057, 28, True),
    'logic_autolock_robust_peak_height_4': (8, 0x05b, 28, True),
    'logic_autolock_robust_peak_height_5': (8, 0x05f, 28, True),
    'logic_autolock_robust_peak_height_6': (8, 0x063, 28, True),
    'logic_autolock_robust_peak_height_7': (8, 0x067, 28, True),
    'logic_autolock_robust_peak_height_8': (8, 0x06b, 28, True),
    'logic_autolock_robust_peak_height_9': (8, 0x06f, 28, True),
    'logic_autolock_robust_peak_height_10': (8, 0x073, 28, True),
    'logic_autolock_robust_peak_height_11': (8, 0x077, 28, True),
    'logic_autolock_robust_peak_height_12': (8, 0x07b, 28, True),
    'logic_autolock_robust_peak_height_13': (8, 0x07f, 28, True),
    'logic_autolock_robust_peak_height_14': (8, 0x083, 28, True),
    'logic_autolock_robust_peak_height_15': (8, 0x087, 28, True),
    'logic_autolock_robust_peak_height_16': (8, 0x08b, 28, True),
    'logic_autolock_robust_peak_height_17': (8, 0x08f, 28, True),
    'logic_autolock_robust_peak_height_18': (8, 0x093, 28, True),
    'logic_autolock_robust_peak_height_19': (8, 0x097, 28, True),
    'logic_autolock_robust_peak_height_20': (8, 0x09b, 28, True),
    'logic_autolock_robust_peak_height_21': (8, 0x09f, 28, True),
    'logic_autolock_robust_peak_height_22': (8, 0x0a3, 28, True),
    'logic_autolock_robust_peak_height_23': (8, 0x0a7, 28, True),
    'logic_autolock_robust_peak_height_24': (8, 0x0ab, 28, True),
    'logic_autolock_robust_peak_height_25': (8, 0x0af, 28, True),
    'logic_autolock_robust_peak_height_26': (8, 0x0b3, 28, True),
    'logic_autolock_robust_peak_height_27': (8, 0x0b7, 28, True),
    'logic_autolock_robust_peak_height_28': (8, 0x0bb, 28, True),
    'logic_autolock_robust_peak_height_29': (8, 0x0bf, 28, True),
    'logic_autolock_robust_peak_height_30': (8, 0x0c3, 28, True),
    'logic_autolock_robust_peak_height_31': (8, 0x0c7, 28, True),
    'logic_autolock_robust_wait_for_0': (8, 0x0cb, 14, True),
    'logic_autolock_robust_wait_for_1': (8, 0x0cd, 14, True),
    'logic_autolock_robust_wait_for_2': (8, 0x0cf, 14, True),
    'logic_autolock_robust_wait_for_3': (8, 0x0d1, 14, True),
    'logic_autolock_robust_wait_for_4': (8, 0x0d3, 14, True),
    'logic_autolock_robust_wait_for_5': (8, 0x0d5, 14, True),
    'logic_autolock_robust_wait_for_6': (8, 0x0d7, 14, True),
    'logic_autolock_robust_wait_for_7': (8, 0x0d9, 14, True),
    'logic_autolock_robust_wait_for_8': (8, 0x0db, 14, True),
    'logic_autolock_robust_wait_for_9': (8, 0x0dd, 14, True),
    'logic_autolock_robust_wait_for_10': (8, 0x0df, 14, True),
    'logic_autolock_robust_wait_for_11': (8, 0x0e1, 14, True),
    'logic_autolock_robust_wait_for_12': (8, 0x0e3, 14, True),
    'logic_autolock_robust_wait_for_13': (8, 0x0e5, 14, True),
    'logic_autolock_robust_wait_for_14': (8, 0x0e7, 14, True),
    'logic_autolock_robust_wait_for_15': (8, 0x0e9, 14, True),
    'logic_autolock_robust_wait_for_16': (8, 0x0eb, 14, True),
    'logic_autolock_robust_wait_for_17': (8, 0x0ed, 14, True),
    'logic_autolock_robust_wait_for_18': (8, 0x0ef, 14, True),
    'logic_autolock_robust_wait_for_19': (8, 0x0f1, 14, True),
    'logic_autolock_robust_wait_for_20': (8, 0x0f3, 14, True),
    'logic_autolock_robust_wait_for_21': (8, 0x0f5, 14, True),
    'logic_autolock_robust_wait_for_22': (8, 0x0f7, 14, True),
    'logic_autolock_robust_wait_for_23': (8, 0x0f9, 14, True),
    'logic_autolock_robust_wait_for_24': (8, 0x0fb, 14, True),
    'logic_autolock_robust_wait_for_25': (8, 0x0fd, 14, True),
    'logic_autolock_robust_wait_for_26': (8, 0x0ff, 14, True),
    'logic_autolock_robust_wait_for_27': (8, 0x101, 14, True),
    'logic_autolock_robust_wait_for_28': (8, 0x103, 14, True),
    'logic_autolock_robust_wait_for_29': (8, 0x105, 14, True),
    'logic_autolock_robust_wait_for_30': (8, 0x107, 14, True),
    'logic_autolock_robust_wait_for_31': (8, 0x109, 14, True),
    'logic_autolock_fast_target_position': (8, 0x10b, 14, True),
    'logic_autolock_request_lock': (8, 0x10d, 1, True),
    'logic_autolock_autolock_mode': (8, 0x10e, 2, True),
    'logic_autolock_lock_running': (8, 0x10f, 1, False),
    'logic_raw_acquisition_iir_z0': (8, 0x110, 27, True),
    'logic_raw_acquisition_iir_a1': (8, 0x114, 25, True),
    'logic_raw_acquisition_iir_a2': (8, 0x118, 25, True),
    'logic_raw_acquisition_iir_a3': (8, 0x11c, 25, True),
    'logic_raw_acquisition_iir_a4': (8, 0x120, 25, True),
    'logic_raw_acquisition_iir_a5': (8, 0x124, 25, True),
    'logic_raw_acquisition_iir_b0': (8, 0x128, 25, True),
    'logic_raw_acquisition_iir_b1': (8, 0x12c, 25, True),
    'logic_raw_acquisition_iir_b2': (8, 0x130, 25, True),
    'logic_raw_acquisition_iir_b3': (8, 0x134, 25, True),
    'logic_raw_acquisition_iir_b4': (8, 0x138, 25, True),
    'logic_raw_acquisition_iir_b5': (8, 0x13c, 25, True),
    'logic_control_signal_clr': (8, 0x140, 1, True),
    'logic_control_signal_max': (8, 0x141, 25, False),
    'logic_control_signal_min': (8, 0x145, 25, False),
    'logic_combined_error_signal_clr': (8, 0x149, 1, True),
    'logic_combined_error_signal_max': (8, 0x14a, 25, False),
    'logic_combined_error_signal_min': (8, 0x14e, 25, False),
    'logic_combined_error_signal_filtered_clr': (8, 0x152, 1, True),
    'logic_combined_error_signal_filtered_max': (8, 0x153, 25, False),
    'logic_combined_error_signal_filtered_min': (8, 0x157, 25, False),
    'scopegen_external_trigger': (6, 0x000, 1, True),
    'scopegen_dac_a_clr': (6, 0x001, 1, True),
    'scopegen_dac_a_max': (6, 0x002, 25, False),
    'scopegen_dac_a_min': (6, 0x006, 25, False),
    'scopegen_dac_b_clr': (6, 0x00a, 1, True),
    'scopegen_dac_b_max': (6, 0x00b, 25, False),
    'scopegen_dac_b_min': (6, 0x00f, 25, False),
    'scopegen_adc_a_sel': (6, 0x013, 4, True),
    'scopegen_adc_b_sel': (6, 0x014, 4, True),
    'scopegen_adc_a_q_sel': (6, 0x015, 4, True),
    'scopegen_adc_b_q_sel': (6, 0x016, 4, True),
    'slow_chain_pid_setpoint': (2, 0x000, 14, True),
    'slow_chain_pid_kp': (2, 0x002, 14, True),
    'slow_chain_pid_ki': (2, 0x004, 14, True),
    'slow_chain_pid_reset': (2, 0x006, 1, True),
    'slow_chain_pid_kd': (2, 0x007, 14, True),
    'slow_chain_limit_min': (2, 0x009, 14, True),
    'slow_chain_limit_max': (2, 0x00b, 14, True),
    'slow_chain_out_clr': (2, 0x00d, 1, True),
    'slow_chain_out_max': (2, 0x00e, 25, False),
    'slow_chain_out_min': (2, 0x012, 25, False),
    'xadc_temp': (29, 0x000, 12, False),
    'xadc_v': (29, 0x002, 12, False),
    'xadc_a': (29, 0x004, 12, False),
    'xadc_b': (29, 0x006, 12, False),
    'xadc_c': (29, 0x008, 12, False),
    'xadc_d': (29, 0x00a, 12, False),
}
states = ['force', 'di0', 'di1', 'di2', 'di3', 'di4', 'di5', 'di6', 'di7', 'robustwatching', 'robustturn_on_lock', 'robustsign_equal', 'robustover_threshold', 'robustwaited_long_enough']
signals = ['zero', 'fast_a_x', 'fast_a_out_i', 'fast_a_out_q', 'fast_b_x', 'fast_b_out_i', 'fast_b_out_q', 'slow_chain_out', 'scopegen_dac_a', 'scopegen_dac_b', 'logic_control_signal', 'logic_combined_error_signal', 'logic_combined_error_signal_filtered']



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import traceback

import numpy as np
from linien_common.common import determine_shift_by_correlation, get_lock_point

from .approach_line import Approacher
from .engine import OptimizerEngine
from .utils import FINAL_ZOOM_FACTOR


class OptimizeSpectroscopy:
    def __init__(self, control, parameters):
        self.control = control
        self.parameters = parameters

        self.initial_spectrum = None
        self.iteration = 0

        self.approacher = None

        self.recenter_after = 2
        self.next_recentering_iteration = self.recenter_after
        self.allow_increase_of_recentering_interval = True

        self.initial_sweep_speed = self.parameters.sweep_speed.value
        self.initial_sweep_amplitude = self.parameters.sweep_amplitude.value
        self.initial_sweep_center = self.parameters.sweep_center.value

    def run(self, x0, x1, spectrum):
        self.parameters.optimization_failed.value = False
        self.parameters.optimization_approaching.value = True

        spectrum = pickle.loads(spectrum)
        cropped = spectrum[x0:x1]
        min_idx = np.argmin(cropped)
        max_idx = np.argmax(cropped)
        self.x0, self.x1 = x0 + min_idx, x0 + max_idx

        self.record_first_error_signal(spectrum)

        params = self.parameters
        self.engine = OptimizerEngine(self.control, params)
        params.to_plot.add_callback(self.react_to_new_spectrum)
        params.optimization_running.value = True
        params.optimization_improvement.value = 0

    def record_first_error_signal(self, error_signal):
        (
            mean_signal,
            _2,
            target_zoom,
            rolled_error_signal,
            line_width,
            peak_idxs,
        ) = get_lock_point(
            error_signal,
            *list(sorted([self.x0, self.x1])),
            final_zoom_factor=FINAL_ZOOM_FACTOR
        )

        self.target_zoom = target_zoom
        self.first_error_signal = rolled_error_signal

        self.approacher = Approacher(
            self.control,
            self.parameters,
            self.first_error_signal,
            self.target_zoom,
            mean_signal,
            allow_sweep_speed_change=False,
        )

    def react_to_new_spectrum(self, spectrum):
        if not self.parameters.optimization_running.value:
            return

        try:
            params = self.parameters

            dual_channel = params.dual_channel.value
            channel = params.optimization_channel.value
            spectrum_idx = 1 if not dual_channel else (1, 2)[channel]
            unpickled = pickle.loads(spectrum)
            spectrum = unpickled["error_signal_%d" % spectrum_idx]
            quadrature = unpickled["error_signal_%d_quadrature" % spectrum_idx]

            if self.parameters.optimization_approaching.value:
                approaching_finished = self.approacher.approach_line(spectrum)
                if approaching_finished:
                    self.parameters.optimization_approaching.value = False
            else:
                self.iteration += 1

                if self.initial_spectrum is None:
                    params = self.parameters
                    self.initial_spectrum = spectrum

                    self.engine.tell(spectrum, quadrature)

                center_line = self.iteration == self.next_recentering_iteration
                center_line_next_time = (
                    self.iteration + 1 == self.next_recentering_iteration
                )

                if self.iteration > 1:
                    if center_line:
                        # center the line again
                        shift, _, _2 = determine_shift_by_correlation(
                            1, self.initial_spectrum, spectrum
                        )
                        params.sweep_center.value -= (
                            shift * params.sweep_amplitude.value
                        )
                        self.control.exposed_write_registers()

                        if (
                            self.allow_increase_of_recentering_interval
                            and abs(shift) < 2 / FINAL_ZOOM_FACTOR
                        ):
                            self.recenter_after *= 2
                        else:
                            self.allow_increase_of_recentering_interval = False

                        self.next_recentering_iteration += self.recenter_after
                    else:
                        self.engine.tell(spectrum, quadrature)

                if not self.engine.finished():
                    self.engine.request_and_set_new_parameters(
                        use_initial_parameters=center_line_next_time
                    )
                else:
                    # we are done!
                    self.exposed_stop(True)

        except Exception:
            print("exception at optimization task")
            traceback.print_exc()
            self.parameters.optimization_failed.value = True
            self.exposed_stop(False)

    def exposed_stop(self, use_new_parameters):
        if use_new_parameters and self.parameters.optimization_improvement.value > 0:
            self.engine.use_best_parameters()
        else:
            self.engine.request_and_set_new_parameters(use_initial_parameters=True)

        self.parameters.optimization_running.value = False
        self.parameters.to_plot.remove_callback(self.react_to_new_spectrum)
        self.parameters.task.value = None

        self.reset_scan()

    def reset_scan(self):
        self.control.exposed_pause_acquisition()

        self.parameters.sweep_speed.value = self.initial_sweep_speed
        self.parameters.sweep_amplitude.value = self.initial_sweep_amplitude
        self.parameters.sweep_center.value = self.initial_sweep_center
        self.control.exposed_write_registers()

        self.control.exposed_continue_acquisition()
